@@ -1,4 +1,4 @@
-# dialogLogic.py - Commit Tool
+# gui_logic.py - Commit Tool
 #
 # Copyright 2006 Steve Borho
 #
@@ -10,7 +10,8 @@ from qctlib.ui_dialog import Ui_commitToolDialog
 from qctlib.ui_preferences import Ui_prefDialog
 from qctlib.utils import formatPatchRichText, runProgram
 from qctlib.version import qct_version
-import os, sys, shutil
+from qctlib.changeselect import ChangeDialog
+import os, sys, shutil, shlex
 
 class CommitTool(QtGui.QWidget):
     '''QCT commit tool GUI logic'''
@@ -52,6 +53,7 @@ class CommitTool(QtGui.QWidget):
             self.ui.commitPushButton.setText(QtGui.QApplication.translate("commitToolDialog",
                 "Refresh Patch", None, QtGui.QApplication.UnicodeUTF8))
             self.commitButtonToolTip = "Refresh selected files in patch " + self.vcs.topPatchName()
+            self.ui.cancelPushButton.setText("Exit")
 
         # Recover persistent data
         settings = QtCore.QSettings('vcs', 'qct')
@@ -66,7 +68,7 @@ class CommitTool(QtGui.QWidget):
         if settings.contains('size'):
             self.resize(settings.value('size').toSize())
             self.move(settings.value('pos').toPoint())
-            self.ui.splitter.restoreState(settings.value('splitter').toByteArray())
+            self.ui.splitter.restoreState(settings.value('3waysplitter').toByteArray())
         settings.endGroup()
         settings.beginGroup('commitLog')
         size = settings.beginReadArray('history')
@@ -285,6 +287,8 @@ class CommitTool(QtGui.QWidget):
                             shutil.rmtree(itemName[2:])
                         else:
                             os.unlink(itemName[2:])
+                    self.__rescanFiles()
+                    self.__refreshFileList(False)
                 elif actionText.startswith('Add all'):
                     selectedFileList = []
                     for item in selectedItemList:
@@ -438,7 +442,7 @@ class CommitTool(QtGui.QWidget):
             menu = QtGui.QMenu()
             if targetFile in self.changeSelectedFiles:
                 menu.addAction("Reset selection of changes")
-            elif targetType == 'M' and self.twowayTool:
+            elif targetType == 'M':
                 menu.addAction("Select changes to commit")
             if targetType not in ['a', 'm', 'r']:
                 menu.addAction("Revert %s to last revisioned state" % targetFile)
@@ -496,7 +500,21 @@ class CommitTool(QtGui.QWidget):
         self.changeSelectedFiles.append(workingFile)
         try:
             self.vcs.generateParentFile(workingFile)
-            runProgram([self.twowayTool, workingCopy, workingFile])
+            if not self.twowayTool:
+                dialog = ChangeDialog(workingCopy, workingFile)
+                dialog.show()
+                dialog.exec_()
+                if not dialog.accepted:
+                    raise Exception
+                return
+            cmd = self.twowayTool
+            if '%o' in cmd and '%m' in cmd:
+                cmd = cmd.replace('%o', workingFile)
+                cmd = cmd.replace('%m', workingCopy)
+                cmd = cmd.replace('\\', '/')
+                runProgram(shlex.split(cmd), expectedexits=[0,1,255])
+            else:
+                runProgram([self.twowayTool, workingCopy, workingFile])
         except:
             print "Change selection failed, returning working file"
             self.__resetChangeSelection(workingFile)
@@ -620,7 +638,7 @@ class CommitTool(QtGui.QWidget):
         settings.beginGroup('mainwindow')
         settings.setValue("size", QtCore.QVariant(self.size()))
         settings.setValue("pos", QtCore.QVariant(self.pos()))
-        settings.setValue("splitter", QtCore.QVariant(self.ui.splitter.saveState()))
+        settings.setValue("3waysplitter", QtCore.QVariant(self.ui.splitter.saveState()))
         settings.setValue("sortby", QtCore.QVariant(self.sortby))
         settings.endGroup()
         settings.beginGroup('commitLog')
@@ -899,16 +917,6 @@ class PrefDialog(QtGui.QDialog):
         self.ui.editToolEdit.setText(self.editTool)
         self.ui.signoffTextEdit.setPlainText(self.signoff)
 
-    def on_mergeToolBrowseButton_pressed(self):
-        fileName = QtGui.QFileDialog.getOpenFileName(self, "Select a two-way merge tool", '', 'All Files (*)')
-        if not fileName.isEmpty():
-            self.ui.mergeToolEdit.setText(fileName)
-
-    def on_editToolBrowseButton_pressed(self):
-        fileName = QtGui.QFileDialog.getOpenFileName(self, "Select an external editor", '', 'All Files (*)')
-        if not fileName.isEmpty():
-            self.ui.editToolEdit.setText(fileName)
-
     def on_aboutPushButton_pressed(self):
         QtGui.QMessageBox.about(self, 'Qct Commit Tool', 
 "<qt><center><h1>Qct " + qct_version + """</h1></center>\n
