@@ -5,32 +5,56 @@
 # This software may be used and distributed according to the terms
 # of the GNU General Public License, incorporated herein by reference.
 
-import os
+import os, sys
 from stat import *
-from qctlib.utils import runProgram, findInSystemPath, scanDiffOutput, isBinary
+from qctlib.utils import *
+from tempfile import mkstemp
+from StringIO import StringIO
+
+def findHg():
+    path = os.environ["PATH"].split(os.pathsep)
+    for d in path:
+        if os.name == "nt":
+            pathexts = os.environ["PATHEXT"].split(os.pathsep)
+            for ext in pathexts:
+                exepath = os.path.join(d, 'hg' + ext)
+                if os.access(exepath, os.X_OK):
+                    try:
+                        runProgram([exepath, 'version'])
+                        return exepath
+                    except:
+                        pass
+        else:
+            exepath = os.path.join(d, 'hg')
+            if os.access(exepath, os.X_OK):
+                try:
+                    runProgram([exepath, 'version'])
+                    return exepath
+                except:
+                    pass
+    return None
 
 class qctVcsHg:
-    def initRepo(self, argv):
+    def initRepo(self, argv, hgdispatch=None):
         '''Initialize your revision control system, open repository'''
 
-        # Find the mercurial binary name
-        if findInSystemPath('hg') != None:
-            self.hg_exe = 'hg'
-        elif findInSystemPath('hg.cmd') != None:
-            self.hg_exe = 'hg.cmd'
-        elif findInSystemPath('hg.bat') != None:
-            self.hg_exe = 'hg.bat'
-        elif findInSystemPath('hg.exe') != None:
-            self.hg_exe = 'hg.exe'
+        self.commitname = None
+
+        # If we're running from the qct.py extension, we can call directly into mercurial
+        if hgdispatch:
+            self.commands = hgdispatch
         else:
-            print "Unable to find hg (.exe, .bat, .cmd) in your path"
-            return -1
+            self.commands = None
+            self.hg_exe = findHg()
+            if not self.hg_exe:
+                print "Unable to find hg (.exe, .bat, .cmd) in your path"
+                return -1
 
         # Verify we have a valid repository
-        output = runProgram([self.hg_exe, 'root'], expectedexits=[0,255])
-        self.repoRoot = output[:-len(os.linesep)]
-        if output.startswith("abort"):
-            print output
+        (out, err) = self.hgcmd(['root'], okresults=[0,255])
+        self.repoRoot = out.split('\n')[0]
+        if err.startswith('abort'):
+            print err
             return -1
 
         if os.getcwd() == self.repoRoot:
@@ -56,25 +80,54 @@ class qctVcsHg:
         # To enable an auto-matic sign-off message:
         # [qct]
         # signoff = Sign-Off: Steve Borho
-        self.signOff = runProgram([self.hg_exe, 'debugconfig', 'qct.signoff'])
+        self.signOff = self.hgcmd(['showconfig', 'qct.signoff'])[0]
 
         # Determine if this repository has any applied Mercurial Queue patches
-        output = runProgram([self.hg_exe, 'qheader'], expectedexits=[0,1,255])
-        recs = output.split(os.linesep)
-        if recs[0] == 'No patches applied':
+        (output, err) = self.hgcmd(['qheader'], okresults=[0,1,255])
+        if err and "unknown command" in err:
             self.isPatchQ = False
             return 0
-        if recs[0] == "hg: unknown command 'qheader'":
+        if output and 'No patches applied' in output:
             self.isPatchQ = False
             return 0
+
+        # Patches only make sense from repository root
+        if not self.runningFromRoot:
+            print "Changing context to repository root: " + self.repoRoot
+            os.chdir(self.repoRoot)
+            self.runningFromRoot = True
+        self.isPatchQ = True
+        return 0
+
+    def hgcmd(self, args, okresults=[0], binary=False):
+        if self.commands:
+            # Use mercurial library directly
+            ostream = StringIO()
+            errstream = StringIO()
+            ret = None
+            try:
+                sys.stdout = ostream
+                sys.stderr = errstream
+                sys.stdin = StringIO()
+                ret = self.commands.dispatch(args)
+            finally:
+                sys.stdin  = sys.__stdin__
+                sys.stdout = sys.__stdout__
+                sys.stderr = sys.__stderr__
+            if ret and ret not in okresults:
+                ret = ret % 256
+                if ret not in okresults:
+                    print "Cmd: hg", ' '.join(args), 'failed with code', ret
+            outstreams = (ostream.getvalue(), errstream.getvalue())
+            ostream.close()
+            errstream.close()
+            return outstreams
         else:
-            # Patches only make sense from repository root
-            if not self.runningFromRoot:
-                print "Changing context to repository root: " + self.repoRoot
-                os.chdir(self.repoRoot)
-                self.runningFromRoot = True
-            self.isPatchQ = True
-            return 0
+            (out, err) = runProgramStderr([self.hg_exe] + args, expectedexits=okresults)
+            if os.linesep != '\n' and not binary:
+                out = out.replace(os.linesep, '\n')
+                err = err.replace(os.linesep, '\n')
+            return (out, err)
 
     def pluginOptions(self, opts):
         '''The Mercurial extension is passing along -I/-X command line options'''
@@ -82,6 +135,9 @@ class qctVcsHg:
             self.cmdLineOptions += ['-X', epath]
         for ipath in opts['include']:
             self.cmdLineOptions += ['-I', ipath]
+        if 'user' in opts:
+            for name in opts['user']:
+                self.commitname = name
 
     def capabilities(self):
         '''Return a list of optional capabilities supported by this VCS'''
@@ -91,32 +147,33 @@ class qctVcsHg:
         '''The GUI needs this file's parent revision in place so the user
            can select individual changes for commit (basically a revert)
         '''
-        runProgram([self.hg_exe, 'cat', '--output', '%d/%s', workingFile])
+        self.hgcmd(['cat', '--output', '%d/%s', workingFile])
 
     def addIgnoreMask(self, newIgnoreMask):
         '''The user has provided a new ignoreMask to be added to revision control'''
-        globString = 'syntax: glob'
+        globString = 'syntax: glob\n'
         # Read existing .hgignore (possibly empty)
         try:
-            f = open(os.path.join(self.repoRoot, '.hgignore'), 'rb')
-            text = f.read()
-            f.close()
-            iLines = text.split(os.linesep)
+            f = open(os.path.join(self.repoRoot, '.hgignore'), 'r')
+            iLines = f.readlines()
         except IOError:
             iLines = []
+        else:
+            f.close()
+
         # Find 'syntax: glob' line, add at end if not found
         if globString in iLines:
             line = iLines.index(globString)
         else:
             iLines.append(globString)
             line = len(iLines) - 1
-            if line == 0:
-                iLines.append('')
+            if line == 0: iLines.append('')
+
         # Insert new mask after 'syntax: glob' line
-        iLines.insert(line + 1, newIgnoreMask)
+        iLines.insert(line + 1, newIgnoreMask + '\n')
         try:
-            f = open(os.path.join(self.repoRoot, '.hgignore'), 'wb')
-            f.write(os.linesep.join(iLines))
+            f = open(os.path.join(self.repoRoot, '.hgignore'), 'w')
+            f.writelines(iLines)
             f.close()
             print "Added '%s' to ignore mask" % newIgnoreMask
         except IOError, e:
@@ -126,13 +183,13 @@ class qctVcsHg:
     def fileMoveDetected(self, origFileName, newFileName):
         '''User has associated an unknown file with a missing file, describing
            a move/rename which occurred outside of revision control'''
-        runProgram([self.hg_exe, 'mv', '--after', origFileName, newFileName])
+        self.hgcmd(['mv', '--after', origFileName, newFileName])
         print "Recording move of %s to %s" % (origFileName, newFileName)
 
     def fileCopyDetected(self, origFileName, newFileName):
         '''User has associated an unknown file with an existing file, describing
            a copy which occurred outside of revision control'''
-        runProgram([self.hg_exe, 'cp', '--after', origFileName, newFileName])
+        self.hgcmd(['cp', '--after', origFileName, newFileName])
         print "Recording copy of %s to %s" % (origFileName, newFileName)
 
     def getLogTemplate(self):
@@ -141,7 +198,8 @@ class qctVcsHg:
         # user is not commiting a changeset. they are refreshing the top patch.
         # So we put the current patch's description in the edit window.
         if self.isPatchQ:
-            return runProgram([self.hg_exe, 'qheader'], expectedexits=[0,1,255])
+            qheader = self.hgcmd(['qheader'], okresults=[0,1,255])[0]
+            return qheader[:-1]
 
         try:
             f = open(os.path.join(self.repoRoot, '.commit.template'), 'r')
@@ -165,13 +223,15 @@ class qctVcsHg:
 
     def topPatchName(self):
         '''Return name of top patch (being refreshed)'''
-        output = runProgram([self.hg_exe, 'qtop'])
+        output = self.hgcmd(['qtop'], okresults=[0,1])[0]
         return output[:-1]
 
     def dirtyCache(self, fileName):
         '''The GUI would like us to forget cached data for this file'''
         if self.wdDiffCache.has_key(fileName):
             del self.wdDiffCache[fileName]
+        if self.patchDiffCache.has_key(fileName):
+            del self.patchDiffCache[fileName]
 
     def scanFiles(self, showIgnored, pb = None):
         '''Request scan for all commitable files from VCS, with optional
@@ -185,21 +245,21 @@ class qctVcsHg:
         # by diff are always relative to the repo root, so if we're running outside of the root
         # directory there is no point in trying to pre-cache diffs.
         if self.runningFromRoot:
-            diff = runProgram([self.hg_exe, 'diff', '--show-function'] + self.cmdLineOptions)
+            diff = self.hgcmd(['diff', '--show-function'] + self.cmdLineOptions)[0]
             (addedList, removedList, modifiedList, self.wdDiffCache) = scanDiffOutput(diff)
 
         if pb: pb.setValue(1)
 
         # Provides ARM, same as diff, plus unknown ? and missing !
-        statusOutput = runProgram([self.hg_exe, 'status'] + self.cmdLineOptions + ['.'])
-        recs = statusOutput.split(os.linesep)
+        statusOutput = self.hgcmd(['status'] + self.cmdLineOptions + ['.'])[0]
+        recs = statusOutput.split('\n')
         recs.pop() # remove last entry (which is '')
 
         if pb: pb.setValue(2)
 
         if showIgnored:
-            statusOutput = runProgram([self.hg_exe, 'status', '-i'] + self.cmdLineOptions + ['.'])
-            recs += statusOutput.split(os.linesep)
+            statusOutput = self.hgcmd(['status', '-i'] + self.cmdLineOptions + ['.'])[0]
+            recs += statusOutput.split('\n')
             recs.pop() # remove last entry (which is '')
 
         if pb: pb.setValue(3)
@@ -207,7 +267,6 @@ class qctVcsHg:
         annotatedFileList = [ ]
         workingDirList = [ ]
         for fileName in recs:
-            if fileName.startswith('*** '): continue   # Skip error messages from hg status
             workingDirList.append(fileName[2:])
             annotatedFileList.append(fileName)
 
@@ -215,8 +274,11 @@ class qctVcsHg:
 
         if self.isPatchQ:
             # Capture changes in the current patch (parse and store hg tip)
-            qtop = runProgram([self.hg_exe, 'tip', '--patch'])
-            (addedPList, removedPList, modifiedPList, self.patchDiffCache) = scanDiffOutput(qtop)
+            modifiedPList = self.hgcmd(['tip', '--debug', '--template', "{files}"])[0].split()
+            addedPList = self.hgcmd(['tip', '--debug', '--template', "{file_adds}"])[0].split()
+            removedPList = self.hgcmd(['tip', '--debug', '--template', "{file_dels}"])[0].split()
+
+            self.filesinpatch = modifiedPList + addedPList + removedPList
 
             # Add patch files which did not show up in `hg status`
             for f in addedPList:
@@ -229,6 +291,22 @@ class qctVcsHg:
         return annotatedFileList
 
 
+    def __getPatchChanges(self, filename, type):
+        if filename not in self.filesinpatch:
+            hgpathname = '/'.join(filename.split(os.sep))
+            if hgpathname not in self.filesinpatch:
+                return '--- Not yet included in patch ---'
+        if self.patchDiffCache.has_key(filename):
+            return self.patchDiffCache[filename]
+        if type in ['A', 'a', 'M', 'm']:
+            self.patchDiffCache[filename] = self.hgcmd(['qdiff', filename])[0]
+            return self.patchDiffCache[filename]
+        if type in ['R', 'r']:
+            self.patchDiffCache[filename] = self.hgcmd(['cat', '-r', '-2', filename])[0]
+            return self.patchDiffCache[filename]
+        return 'unknown patch state!'
+
+
     def __getWorkingDirChanges(self, fileName, type):
         if self.wdDiffCache.has_key(fileName):
             return self.wdDiffCache[fileName]
@@ -243,7 +321,7 @@ class qctVcsHg:
 
         # For revisioned files, we use hg diff
         if type in ['A', 'M', 'R']:
-            text = runProgram([self.hg_exe, 'diff', '--show-function', fileName])
+            text = self.hgcmd(['diff', '--show-function', fileName])[0]
             self.wdDiffCache[fileName] = text
             return text
 
@@ -260,7 +338,7 @@ class qctVcsHg:
 
         # For missing files, we use hg cat
         if type == '!':
-            text = runProgram([self.hg_exe, 'cat', fileName])
+            text = self.hgcmd(['cat', fileName], binary=True)[0]
             if not text: text = " <empty file>"
             elif '\0' in text: text = " <Binary File of %d KBytes>" % (len(text) / 1024)
             self.wdDiffCache[fileName] = text
@@ -278,30 +356,32 @@ class qctVcsHg:
         fileName = itemName[2:]
         text = self.__getWorkingDirChanges(fileName, type)
 
+        linesep = '\n'
+
         # Useful shorthand vars.  Leading lines beginning with % are treated as RTF
         bFileName = "%<b>" + fileName + "</b>"
-        noteLineSep = os.linesep + '%'
+        noteLineSep = linesep + '%'
 
         if type == 'A':
             note = bFileName + " has been added to revision control or is a rename target, but has never been commited."
-            return note + os.linesep + text
+            return note + linesep + text
         elif type == 'M':
             note = bFileName + " has been modified in your working directory."
-            return note + os.linesep + text
+            return note + linesep + text
         elif type == '?':
             note = bFileName + " is not currently tracked. If commited, it will be added to revision control."
-            return note + os.linesep + "= Unrevisioned File Contents" + os.linesep + text
+            return note + linesep + "= Unrevisioned File Contents" + linesep + text
         elif type == 'I':
             note = bFileName + " is usually ignored, but will be added to revision control if commited"
-            return note + os.linesep + text
+            return note + linesep + text
         elif type == 'R':
             note = bFileName + " has been marked for deletion, or renamed, but has not yet been commited"
             note += noteLineSep + "The file can be recovered by reverting it to it's last revisioned state."
-            return note + os.linesep + "= Removed File Diffs" + os.linesep + text
+            return note + linesep + "= Removed File Diffs" + linesep + text
         elif type == '!':
             note = bFileName + " was tracked but is now missing. If commited, it will be marked as removed in revision control."
             note += noteLineSep + "The file can be recovered by reverting it to it's last revisioned state."
-            return note + os.linesep + "= Contents of Missing File" + os.linesep + text
+            return note + linesep + "= Contents of Missing File" + linesep + text
         else:
             return "Unknown file type " + type
 
@@ -314,64 +394,65 @@ class qctVcsHg:
 
         # Useful shorthand vars.  Leading lines beginning with % are treated as RTF
         bFileName = "%<b>" + fileName + "</b>"
-        noteLineSep = os.linesep + '%'
+        linesep = '\n'
+        noteLineSep = linesep + '%'
 
         if type == 'A':
             note = bFileName + " has been added to the working directory, but has not been included in this patch."
             note += noteLineSep + "If reverted, this file will return to an unrevisioned state."
             wtext = self.__getWorkingDirChanges(fileName, type)
-            return note + os.linesep + "= Added File Diffs" + os.linesep + wtext
+            return note + linesep + "= Added File Diffs" + linesep + wtext
         elif type == 'a':
             note = bFileName + " is a new file provided by this patch.  "
             note += noteLineSep + "Reverting this file has no effect, it must be removed from the patch first."
-            ptext = self.patchDiffCache[fileName]
-            return note + os.linesep + "= Added File Diffs" + os.linesep + ptext
+            ptext = self.__getPatchChanges(fileName, type)
+            return note + linesep + "= Added File Diffs" + linesep + ptext
         elif type == '?':
             note = bFileName + " is not currently tracked. If commited, it will appear to be provided by this patch.  "
             note += noteLineSep + "Reverting this file has no effect."
             wtext = self.__getWorkingDirChanges(fileName, type)
-            return note + os.linesep + "= Unrevisioned File Contents" + os.linesep + wtext
+            return note + linesep + "= Unrevisioned File Contents" + linesep + wtext
         elif type == 'I':
             note = bFileName + " is usually ignored, but will be recorded as provided by this patch if commited.  "
             note += noteLineSep + "Reverting this file has no effect."
             wtext = self.__getWorkingDirChanges(fileName, type)
-            return note + os.linesep + "= Unrevisioned File Contents" + os.linesep + wtext
+            return note + linesep + "= Unrevisioned File Contents" + linesep + wtext
         elif type == '!':
             note = bFileName + " was tracked but is now missing, will be marked as removed by this patch if commited.  "
             note += noteLineSep + "If reverted, this file will be recovered to last revisioned state."
             wtext = self.__getWorkingDirChanges(fileName, type)
-            return note + os.linesep + "= Contents of Missing File" + os.linesep + wtext
+            return note + linesep + "= Contents of Missing File" + linesep + wtext
         elif type == 'R':
             note = bFileName + " has been marked for deletion in your working directory, but has not yet been commited.  "
             note += noteLineSep + "If reverted, this file will be recovered to it's last revisioned state."
             wtext = self.__getWorkingDirChanges(fileName, type)
-            return note + os.linesep + "= Removed File Diffs" + os.linesep + wtext
+            return note + linesep + "= Removed File Diffs" + linesep + wtext
         elif type == 'r':
             note = bFileName + " is deleted by this patch"
             note += noteLineSep + "If you remove this file from the patch, it will appear removed in your working dir, "
             note += noteLineSep + "at which point you can revert it to it's last revisioned state."
-            ptext = self.patchDiffCache[fileName]
-            return note + os.linesep + "= Removed File Diffs" + os.linesep + ptext
+            ptext = self.__getPatchChanges(fileName, type)
+            return note + linesep + "= Removed File Diffs" + linesep + ptext
         elif type == 'M':
             wtext = self.__getWorkingDirChanges(fileName, type)
-            if self.patchDiffCache.has_key(fileName):
+            ptext = self.__getPatchChanges(fileName, type)
+            if ptext:
                 note = bFileName + " has changes recorded in the patch, and further changes in the working directory "
                 note += noteLineSep + "If reverted, only the working directory changes will be removed.  "
                 note += noteLineSep + "If you refresh without this file, all changes will be left in your working directory."
-                ptext = self.patchDiffCache[fileName]
-                status = note + os.linesep + "= Working Directory Diffs" + os.linesep + wtext
-                status += os.linesep + "= Patch Diffs" + os.linesep + ptext
+                status = note + linesep + "= Working Directory Diffs" + linesep + wtext
+                status += linesep + "= Patch Diffs" + linesep + ptext
             else:
                 note = bFileName + " has changes in the working directory that are not yet included in this patch."
                 note += noteLineSep + "If reverted, the working directory diffs will be removed."
-                status = note + os.linesep + "= Working Directory Diffs" + os.linesep + wtext
+                status = note + linesep + "= Working Directory Diffs" + linesep + wtext
             return status
         elif type == 'm':
             note = bFileName + " is modified by this patch.  There are no further changes in the working directory so "
             note += noteLineSep + "reverting this file will have no effect.  If you remove this file from the patch "
             note += noteLineSep + "these modifications will be left in the working directory."
-            ptext = self.patchDiffCache[fileName]
-            return note + os.linesep + "= Patch Diffs" + os.linesep + ptext
+            ptext = self.__getPatchChanges(fileName, type)
+            return note + linesep + "= Patch Diffs" + linesep + ptext
         else:
             return "Unknown file type " + type
 
@@ -391,27 +472,37 @@ class qctVcsHg:
             elif type == '!': removeFileList.append(fileName)
 
         if addFileList:
-            runProgram([self.hg_exe, 'add'] + addFileList)
+            self.hgcmd(['add'] + addFileList)
             print "Added %d file(s) to revision control: %s" % (len(addFileList), ', '.join(addFileList))
 
         if removeFileList:
-            runProgram([self.hg_exe, 'rm'] + removeFileList)
+            self.hgcmd(['rm'] + removeFileList)
             print "Removed %d file(s) from revision control: %s" % (len(removeFileList), ', '.join(removeFileList))
 
         if self.signOff:
-            logMsgText += os.linesep + self.signOff
+            logMsgText += '\n' + self.signOff
+
+        (fd, filename) = mkstemp()
+        file = os.fdopen(fd, "w+b")
+        file.write(logMsgText)
+        file.close()
 
         if self.isPatchQ:
-            runProgram([self.hg_exe, 'qrefresh', '-l', '-'] + commitFileNames, logMsgText)
+            self.hgcmd(['qrefresh', '-l', filename] + commitFileNames)
             print self.topPatchName() + " refreshed with %d file(s): %s" \
                     % (len(commitFileNames), ', '.join(commitFileNames))
         else:
-            runProgram([self.hg_exe, 'commit', '-l', '-'] + commitFileNames, logMsgText)
+            if self.commitname:
+                self.hgcmd(['commit', '-u', self.commitname, '-l', filename] + commitFileNames)
+            else:
+                self.hgcmd(['commit', '-l', filename] + commitFileNames)
             print "%d file(s) commited: %s" % (len(commitFileNames), ', '.join(commitFileNames))
 
+        os.unlink(filename)
+
     def addFiles(self, selectedFileList):
         '''Add selected files to version control'''
-        runProgram([self.hg_exe, 'add'] + selectedFileList)
+        self.hgcmd(['add'] + selectedFileList)
 
     def revertFiles(self, selectedFileList):
         '''Revert selected files to last revisioned state'''
@@ -430,7 +521,7 @@ class qctVcsHg:
                 print "File %s not reverted" % fileName
 
         if revertFileNames:
-            runProgram([self.hg_exe, 'revert'] + revertFileNames)
+            self.hgcmd(['revert'] + revertFileNames)
         else:
             print "No revertable files"
 
